"""
Paypal2

:copyright: 2024 Nándor Mátravölgyi
:license: Proprietary, see LICENSE for more details.
"""

__copyright__ = "Copyright 2024 Nándor Mátravölgyi"
__credits__ = ["Nándor Mátravölgyi"]
__version__ = "0.5.0"
